:- module( hdrug_gui,
	   [ start_x/0,  really_start_x/0,  restart_x/0,  halt_x/0, 
	     update_types/0, update_preds/0, update_sents/0, update_lfs/0,
	     update_array/2, update_array/3,
	     tk_retract_objects/1, tk_assert_object/1,
	     tk_fs/1, tk_fs_list/1, tk_term/1, tk_term_list/1, tk_tree/2,
	     tcl_eval/1, tcl_eval/2, tcl_eval/3, tcl/1, tcl/2, tcl/3,
	     tk_num_main_windows/1,
	     tk_atom/2,
	     menu_flag/1, menu_flag/2,
	     notify_active_obj/1,
	     select_widget/2,
	     reposition_widget/1,
	     create_button_for_object/1,
	     create_button_for_object/2,
	     create_more_buttons/1,
	     create_less_buttons/1,
	     select_canvas/2,  %% obsolete, use select_widget/2 instead
	     tkconsol_larger/0,
	     tkconsol_larger/1,
	     tkconsol_smaller/0,
	     tkconsol_smaller/1,
	     gui_user_confirmation/2
	   ]).

%% Hdrug Graphical User Interface


:- multifile user:help_pred/3.
:- public user:help_pred/3.

:- multifile user:help_hook/3.
:- public user:help_hook/3.

:- meta_predicate findall_atmost(?,?,:,?), findall_atmost_(?,?,:,?,?).

:- use_module( library(lists)).
:- use_module( hdrug_util ).
:- use_module( library(charsio)).
:- use_module( library(tcltk)).

%% starting and halting tk interface
% x starts up Tk environment, unless the option `notk' is present.

:- initialize_flag(tcltk,on).
:- initialize_flag(complex_parse_widget,on).

:- public help_pred/3.

help_pred(really_start_x,"really_start_x",
"Starts the graphical user interface, even if the tcltk flag is
switched off."). 
really_start_x :-
    set_flag(tcltk,on),
    restart_x.

help_pred(restart_x,"restart_x","Restarts the graphical user interface.").
restart_x :-
    halt_x,
    start_x.

load_blt :-
    user:blt_lib(BLT,LIB),
    load_blt(BLT,LIB).

load_blt(noblt,_).
load_blt(blt,BltLib) :-
    (	BltLib \== ''
    ->	on_exception(tcl_error(_,ErrorString),
		     (	 
			 tcl('load ~a',[BltLib]),
			 format(user_error,"BLT extension loaded~n",[])
		     ),
		     (	 format(user_error,"Could not load BLT: ~s~n",
				[ErrorString]),
			 format(user_error,"Maybe old Tcl/Tk version?~n",[])
		     ))
    ;	format(user_error,"BLT library not specified, ",[]),
	format(user_error,"BLT extension not loaded ~n",[]),
	format(user_error,
	       "(but maybe it's already loaded statically)~n",[])
    ). 

user:help_pred(start_x,"start_x",
"Attempts to start the graphical user interface, but will not start it
if flag(tcltk) is switched off"). 
start_x :-
    flag(tcltk,TcltkFlag),
    (	TcltkFlag == on
    ->	tk_init,
	load_blt,
	%% enables hdrug library
	user:hdrug_library(Lib),
	tcl("set auto_path [linsert $auto_path 0 ~a]",[Lib]),
	tcl("set hdrug_library ~a",[Lib]),
	tcl("set autopath [linsert $auto_path 0 ~a/Clig]",[Lib]),
	tcl("set env(CLIG_DIR) ~a/Clig",[Lib]),
	tcl("source ~a/Clig/main.tcl",[Lib]),
	tcl("set background_color wheat3"),
	tcl("set highlight_color DarkSeaGreen1"),

	bb_put(user:hdrug_gui,started),

	%% get information about current grammar
	%% It would be too slow to ask for this dynamically.
	update_flags,
	update_types,
	update_preds,
	update_sents,
	update_lfs,
	flag(application_name,Val),
	(   ( Val==undefined
	    ;	\+ atom(Val)
	    )
	->  AppName='Hdrug'
	;   AppName=Val
	),
	tcl('set HDRUG ~a',[AppName]),
	try_hook(user:gram_startup_hook_begin),
	tcl_eval('source $hdrug_library/hdrug.tcl'),
	hdrug_menu_flag( useful_try_check,    [ on, off ]),
	hdrug_menu_flag( nodeskip,            [ 0,1,2,3,4,5,6,7,8,9,10]),
	hdrug_menu_flag( demo,                [ on, off ]),
	hdrug_menu_flag( object_saving,       [ on, off, semi ]),
	hdrug_menu_flag( object_exists_check, [ on, off ]),
	hdrug_menu_flag( max_objects,         [ 1,10,40,100,1000000]),
	hdrug_menu_flag( gui_max_objects,     [ 30,40,50,60 ]),
	hdrug_menu_flag( application_name ),
	hdrug_menu_flag( blt_graph_lines,     [ on, off]),
	hdrug_menu_flag( debug,               [ 0, 1, 2]),
	hdrug_menu_flag( clig_tree_hspace,    [3,5,10,20,30,40,50 ]),
	hdrug_menu_flag( clig_tree_vspace,    [3,5,10,20,30,40,50 ]),
	hdrug_menu_flag( clig_tree_active_nodes, [ on,off]),
	hdrug_menu_flag( print_table_total,   [ on, off]),
	hdrug_menu_flag( start_results_within_bound ),
	hdrug_menu_flag( end_results_within_bound ),
	hdrug_menu_flag( incr_results_within_bound ),
	hdrug_menu_flag( complex_parse_widget, [on, off]),
	hdrug_menu_flag( tcltk ),
	hdrug_menu_flag( tkconsol,            [ on, off] ),
	tcl("create_objects",[]),
	flag(tcl_interp,Interp),
	flag(add_help_menu,On),
	(   On==off
	->  true
	;   help:help_add_to_menu('.menu.help.m',Interp)
	),
	start_tkconsol,
	try_hook(user:gram_startup_hook_end),
	tk_main_loop_if_no_top_level_events
    ;	true
    ).

%% Windows systems do not support top_level events (i.e. allowing
%% tcltk to interpret events while SICStus is waiting for input).
tk_main_loop_if_no_top_level_events :-
    prolog_flag(host_type,Host),
    (   sub_atom(Host,_,_,_,win)
    ->  add_toplevel,
        tcltk:tk_main_loop
    ;   true
    ).

add_toplevel :-
    tcl("\
button .t.prolog -text {Prolog Command} -command {
    prolog hdrug_runtime_cmd_interpreter }
pack .t.prolog -side right",[]).   

help_pred(halt_x,"halt_x",
"Halts the graphical user interface; Hdrug continues.").
halt_x :-
    if_gui((reset_tkconsol,
	    tcl_eval('destroy .'),
	    bb_put(user:hdrug_gui,not_started)
	   )).

update_flags :-
    (	flag(X),
	tk_flag(X),
	fail
    ;   true
    ).

update_types :-
    format(user_error,"Updating types...",[]),
    tcl("set types(max) 0"),
    findall(Type,hook(hdrug_feature:define_type(Type,_,_,_,_)),All0),
    sort(All0,All),
    update_array(All,types),
    format(user_error,"Done.~n",[]).

update_preds :-
    format(user_error,"Updating preds...",[]),
    tcl("set preds(max) 0"),
    findall(Spec, user_clause_functor(Spec), All0),
    sort(All0,All),
    update_array(All,preds),
    format(user_error,"Done.~n",[]).

user_clause_functor(F/A):-
    hook(user:user_clause(H,_)),
    functor(H,F,A).

user:help_pred(update_array,"update_array(+List,+ArrayName)",
"a Tcl array named ArrayName is constructed where the values in List
are to be the values in the array, i.e. ArrayName(1), ArrayName(2),
etc.; the special value ArrayName(max) is set to the last index of the
array (counting starts at 0). The flag update_array_max can be used to
pass to Tcl only the first N items. If that value is 0 then all items
are passed on (default=1000)."). 

:- initialize_flag(update_array_max,1000).

update_array(Values0,Var) :-
    flag(update_array_max,Max),
    atmost_prefix(Max,Values0,Values),
    update_array(Values,Var,0).

%update_array(Values,Var,Int)
update_array([],Var,Int) :-
    tcl("set ~w(max) ~w",[Var,Int]).
update_array([H0|T],Var,Int) :-
    term_atom(H0,H),
    tcl("set ~w(~w) {~w}",[Var,Int,H]),
    Int2 is Int+1,
    update_array(T,Var,Int2).

update_sents :-
    flag(complex_parse_widget,OnOff),
    update_sents(OnOff).

update_sents(off).
update_sents(on) :-
    format(user_error,"Updating sentences...",[]),
    tcl("set sents(max) 0"),
    flag(update_array_max,Max),
    findall_atmost(Max,S,a_sentence(S),Ss0),
    to_tk_sentences(Ss0,Ss),
    update_array(Ss,sents),
    format(user_error,"Done.~n",[]).

update_lfs :-
    format(user_error,"Updating logical forms...",[]),
    tcl("set lfs(max) 0"),
    flag(update_array_max,Max),
    findall_atmost(Max,S,a_lf(S),Ss0),
    to_tk_lfs(Ss0,Ss),
    update_array(Ss,lfs),
    format(user_error,"Done.~n",[]).

a_lf(Lf) :-
    hook(user:lf(_,Lf)).

to_tk_lfs([],[]).
to_tk_lfs([Lf0|Lfs0],[Lf|Lfs]) :-
    try_hook(user:extern_sem(Lf0,Lf),Lf0=Lf),
    to_tk_lfs(Lfs0,Lfs).

a_sentence(S) :-
    user:a_sentence(_,S).

to_tk_sentences([],[]).
to_tk_sentences([H0|T0],[H|T]) :-
    to_tk_sentence(H0,H),
    to_tk_sentences(T0,T).

to_tk_sentence(S0,S) :-
    create_list_of_atoms(S0,Chars,[]),
    atom_chars(S,Chars).

create_list_of_atoms([],C,C).
create_list_of_atoms([H|T],C0,C) :-
    create_list_of_atoms_(T,H,C0,C).

create_list_of_atoms_([],H,C0,C) :-
    charsio:format_to_chars("~w",[H],C0,C).
create_list_of_atoms_([H|T],F,C0,C) :-
    charsio:format_to_chars("~w ",[F],C0,C1),
    create_list_of_atoms_(T,H,C1,C).


%% predicates called from Prolog
%%
%% i.e. these should succeed if tk interface is not started.

%% these predicates are used in go.pl
%% tk_retract_objects/1
%% tk_assert_object/1
%% tk_send_a_sentence/1
%% tk_send_a_lf/1

tk_retract_objects(_) :-
    if_gui(tk_retract_objects0).

tk_retract_objects0 :-
    tcl("hdrug_delete_objects",[]).

tk_assert_object(No) :-
    if_gui(tk_assert_object0(No),
	   format(user_error,"created object: ~w~n",[No])).

tk_assert_object0(No) :-
    flag(demo,OnOff),
    tk_assert_object1(OnOff,No).

tk_assert_object1(off,_).
tk_assert_object1(on,No):-
    create_button_for_object(No).

:- initialize_flag(gui_max_objects,30).

create_button_for_object(No) :-
    flag(gui_max_objects,Max),
    create_button_for_object(No,Max).
    
create_button_for_object(No,Max) :-
    tcl("set flag(number_of_objects) ~w",[No]),
    (	No > Max
    ->	true
    ;	No =:= Max
    ->	tcl("catch {destroy .bb.o.obj~w}
	     button .bb.o.obj~w -padx 0.5 -pady 0.5 -text {>} -command {
	              prolog create_more_buttons(~w) }
             pack .bb.o.obj~w -side left",[Max,Max,Max,Max]),
	tcl('set add_object 0
catch { destroy .bb.text .bb.label .bb.omax .bb.omaxv }
label .bb.label -text {add button for object:}
label .bb.omax -text {objects}
label .bb.omaxv -textvariable flag(number_of_objects) -width 8
entry .bb.text -textvariable add_object -width 10
pack .bb.omax .bb.omaxv .bb.text .bb.label -side right
bind .bb.text <Return> {if [prolog object($add_object,_)] { hdrug_object $add_object} }',[])
    ;   create_button_for_any_object(No)
    ),
    tcl(update).

create_button_for_any_object(No) :-
    tcl("hdrug_object ~w",[No]).

create_more_buttons(OldMax) :-
    flag(gui_max_objects,Inc),
    NewMax is OldMax+Inc,
    tcl("hdrug_delete_objects",[]),
    tcl("button .bb.o.obj1  -padx 0.5 -pady 0.5 -text {<} -command {
                    prolog create_less_buttons(~w) }",[OldMax]),
    tcl("pack .bb.o.obj1 -side left",[]),
    (	between(OldMax,NewMax,N),
	user:object(N,_),
	create_button_for_object(N,NewMax),
	fail
    ;	true
    ).


create_less_buttons(OldMax) :-
    flag(gui_max_objects,Inc),
    NewMax is OldMax-Inc,
    tcl("hdrug_delete_objects",[]),
    (	NewMax > 0
    ->	tcl("button .bb.o.obj1  -padx 0.5 -pady 0.5 -text {<} -command { 
                        prolog create_less_buttons(~w) }",[NewMax]),
	tcl("pack .bb.o.obj1 -side left",[])
    ;	true
    ),
    (	between(NewMax,OldMax,N),
	user:object(N,_),
	create_button_for_object(N,OldMax),
	fail
    ;	true
    ).


%% treatment of canvases
%% there are two `standard' canvases, and an infinite number of new
%% toplevel canvases.
%% 
%% the name of the standard canvases is available thru the value of
%% the flag(canvas) and flag(canvas2)
%%
%% the current canvas will be the value of canvas passed on to the
%% different libraries. In the libraries I don't want to see those
%% flags predicates anymore (at best a local one ..)
%%
%% for each type of output, one of the standard canvases is used as
%% the default. The boolean value of flag(use_default_canvas) indicates
%% whether the default canvas or a new one should be used.

create_top_canvas(Canvas) :-
    create_top_canvas(Canvas,_).

create_top_canvas(Canvas,W) :-
    gen_sym(W,'.cs'),
    tcl("create_top_canvas ~a",[W],Canvas).

user:help_pred(tk_fs,"tk_fs(+Term)",
"Term is displayed as a feature-structure on the canvas widget of the
graphical user interface"). 
user:help_pred(tk_fs_list,"tk_fs(List)",
"Each Term in List is displayed as a feature-structure on the canvas
widget of the graphical user interface"). 
tk_fs(Thing) :-
    if_gui( (   select_widget(fs,Frame),
		user:tk_fs(Thing,Frame),
		reposition_widget(fs)
	    )
	  ).

%% Nov 23, 2000: reverse list, since tk is used as a stack.
tk_fs_list(List0) :-
    lists:reverse(List0,List),
    tk_fs_list_(List).

tk_fs_list_([]).
tk_fs_list_([H|T]) :-
    tk_fs(H),
    tk_fs_list_(T).

%%% this is really a mess: we are having a text-widget here, but
%%% our scrollbars only send messages to the canvas. Solution:
%%% make a very large text widget .... However, now it is not so clear
%%% how to automatically `watch' the insertion cursor.
%%%
%%% other possible solution: don't use text widget at all, but use tk_fs
%%%    currently tk_fs is not very good at terms yet...
user:help_pred(tk_term,"tk_term(?Term)","Term is displayed on the canvas of the graphical user interface").
tk_term(Thing) :-
    if_gui( (   select_widget(text,Text),
		tcl("text ~a -wrap char -width 3000 -height 10 -bd 0",[Text]),
		user:tk_term(Thing,Text),
		reposition_widget(text)
	    )
	  ).  %% -height 1500 

tk_term_list(Thing) :-
    if_gui( (   select_widget(text,Text),
		tcl("text ~a -wrap char -width 3000 -height 10 -bd 0",[Text]),
		user:tk_term_list(Thing,Text),
		reposition_widget(text)
	   )
	 ).

help_pred(tk_tree,"tk_tree(+TreeFormat,?Term)",
"Term is displayed as a tree according to the TreeFormat declarations
on the canvas of the graphical user interface. The TreeFormat should
be specified by means of clauses for the hook predicates graphic_path,
graphic_label, graphic_daughter."). 
tk_tree(Mode,FS) :-
    if_gui(  (   select_widget(tree(Mode),Frame),
		 user:tk_tree(Mode,FS,Frame),
		 reposition_widget(tree(Mode))
	     )
	  ).

canvas_frame(Canvas,Frame) :-
    flag(grr,N0),
    N is N0+1,
    set_flag(grr,N),
    charsio:format_to_chars('~w.f.f~w.f',[Canvas,N0],FrameChars),
    atom_codes(Frame,FrameChars),
    tcl('winfo exists ~w.f',[Canvas],Return),
    (	Return=='0'
    ->	tcl('
	   empty_canvas ~a
	   frame ~a.f
	   pack ~a.f -side left -anchor n -expand 1 -fill x
	   ~a create window 20 5 -window ~a.f -anchor nw
           ~a configure -width [lindex [.cv.canvas configure -width] 4]
           ~a configure -height [lindex [.cv.canvas configure -height] 4]
	   ',[Canvas,Canvas,Canvas,Canvas,Canvas,Canvas,Canvas])
    ;   true
    ),
    tcl('
       frame ~w.f.f~w
       frame ~w.f.g~w -height 40
       pack ~w.f.g~w ~w.f.f~w -side bottom -anchor nw -expand 0
       catch {destroy ~w}
       ',
       [Canvas,N0,Canvas,N0,Canvas,N0,Canvas,N0,Frame]).
    
select_widget(Type,Widget) :-
    select_canvas(Type,Canvas),
    canvas_frame(Canvas,Widget).

reposition_widget(Type) :-
    tcl("global use_canvas ; set use_canvas",[],Return),
    (	Return == '1'
    ->	true
    ;	select_standard_canvas(Type,Canvas),
	tcl('~w yview moveto 0',[Canvas]),
	tcl('~w xview moveto 0',[Canvas])
    ).

select_canvas(Type,Canvas) :-
    tcl("global use_canvas ; set use_canvas",[],Return),
    (	Return == '1'
    ->	create_top_canvas(Canvas)
    ;	select_standard_canvas(Type,Canvas)
    ).

user:help_hook(use_canvas,"use_canvas(+Mode,LeftRightTop)",
"Mode is a term indicating the type of data-structure to be displayed. It
is one of tree(TreeMode), fs, text, chart, stat. The predicate should 
instantiate the second argument as one of the atoms left, right or top 
(for a new widget).").

select_standard_canvas(Mode,Canvas) :-
    try_hook(user:use_canvas(Mode,Val),Val=system),
    select_standard_canvas0(Val,Mode,Canvas).

select_standard_canvas0(top,_,Canvas) :-
    create_top_canvas(Canvas).
select_standard_canvas0(left,_,Canvas) :-
    flag(canvas,Canvas).
select_standard_canvas0(right,_,Canvas) :-
    flag(canvas2,Canvas).
select_standard_canvas0(system,tree(_Mode),Canvas):-
    flag(canvas,Canvas).
select_standard_canvas0(system,fs,Canvas):-
    flag(canvas2,Canvas).
select_standard_canvas0(system,text,Canvas) :-
    flag(canvas2,Canvas).
select_standard_canvas0(system,stat,Canvas) :-
    create_top_canvas(Canvas).
select_standard_canvas0(system,chart,Canvas) :-
    create_top_canvas(Canvas).

:- public a_treedef/1.  % from hdrug.tcl
a_treedef(TreeDef) :-
    findall(G,hook(user:graphic_path(G,_,_)),Gs0),
    sort(Gs0,Gs),
    member(TreeDef,Gs),
    TreeDef \== type,    %% hdrug(p_type), not for objects
    TreeDef \== t.       %% hdrug(call_tree), not for objects

%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% utilities %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%

%tcl_delete :-
%	tcl_interp(I),
%	tcl_delete(I).
%
%tcl_new :-
%	tcl_new(I),
%	set_flag(tcl_interp,I).

user:help_pred(tcl_eval,"tcl_eval(+Cmd[,-Return])",
"Abbrevation for the tcltk library predicate tcl_eval/3. The current
TclTk interpreter, accessible through the tcl_interp flag, is added as
the first argument."). 
tcl_eval(Cmd) :-
	tcl_interp(I),
	tcl_eval(I,Cmd,_).
tcl_eval(Cmd,Return) :-
	tcl_interp(I),
	tcl_eval(I,Cmd,Return).

user:help_pred(tcl,"tcl(+Expr[,+Subs[,-ReturnAtom]])",
"Expr is a string as accepted as the second argument of format/3; the
optional Subs is equivalent to the third argument of format/3. After
evaluating the meta-charcters in Expr, the string is sent as a tcl
command using the current tcl interpreter (flag tcl_interp). The
return string is turned into an atom and available in the optional
third argument."). 
tcl(Expr) :-
    tcl(Expr,[],_).

tcl(Expr,Subs) :-
    tcl(Expr,Subs,_).

tcl(Expr,Subs,Return) :-
    tcl_interp(Interp),
    tcl_eval(Interp,format(Expr,Subs),Return0),
    atom_chars(Return,Return0).

tk_new(Options) :-
    flag(tcl_interp,I),
    (	I \== undefined
    ->	tcl_delete(I)
    ;	true
    ),
    tk_new(Options,Int),
    flag(tcl_interp,_,Int).

%tk_main_window(W) :-
%    tcl_interp(Int),
%    tk_main_window(Int,W).

%% tk_atom(Term,Atom)
%% converts between Prolog term and atom with appropriate
%% symbols escaped for Tcl/Tk 

% don't trust this!
tk_atom(A,B) :-
	integer(A),!,
	number_chars(A,Astr),
	atom_chars(B,Astr).
tk_atom(A0,A) :-
	term_atom(A0,A1),
	name(A1,Chars1),
	tk_escape_l(Chars1,Chars),!,
	name(A,Chars).

tk_escape_l([],[]).
tk_escape_l([H|T],Out):-
	tk_escape_c(H,Rest,Out),
	tk_escape_l(T,Rest).

% 92: \

% 59: ;
% 32: <SPACE>
% 34: "
% 91: [
% 93: ]
% 123 {
% 125 }
% 36: $
tk_escape_c(34,R,[92,34|R]).
tk_escape_c(59,R,[92,59|R]).
tk_escape_c(91,[93|R],[92,91,92,32,92,93|R]).
tk_escape_c(91,R,[92,91|R]).
tk_escape_c(92,R,[92,92|R]).
tk_escape_c(93,R,[92,93|R]).
tk_escape_c(32,R,[92,32|R]).
tk_escape_c(123,R,[92,123|R]).
tk_escape_c(125,R,[92,125|R]).
tk_escape_c(36,R,[92,36|R]).

% catch all:
tk_escape_c(C,R,[C|R]).

%escape_double_quote([],[]).
%escape_double_quote([H|T],Rest) :-
%    escape_dq(H,Rest,Rest1),
%    escape_double_quote(T,Rest1).
%
%escape_dq(C,Rest,Rest1) :-
%    (	C =:= 34
%    ->  Rest=[92,34|Rest1]
%    ;	C =:= 123
%    ->	Rest=[92,123|Rest1]
%    ;	C =:= 125
%    ->	Rest=[92,125|Rest1]
%    ;	C =:= 36
%    ->	Rest=[92,36|Rest1]
%    ;	Rest=[C|Rest1]
%    ).

%%%%%%%%%%%%%%%%%%%%%%
%% Global variables %%
%%%%%%%%%%%%%%%%%%%%%%

% sets the flag Flag, and passes the flag to Tcl/Tk
% flag is supposedly ground
tk_flag(Flag) :-
    flag(Flag,Val),
    if_gui(tcl("set flag(~q) {~q}",[Flag,Val])). % then send it to tk

% menu_flag(Flag,Values)
% adds a radiobutton for Flag under the Options menu, with Values
% an instantiated list of possible values
:- use_module(library(lists)).

menu_flag(Flag) :-
    menu_flag(Flag,[]).

menu_flag(Flag,Values0) :-
	flag(Flag,Value),
	(   member(Value,Values0)
	->  Values0=Values
	;   Values=[Value|Values0]
	),
	tcl("set flag(~w) ~w",[Flag,Value]),
	menu_flag('.menu.options.m',Flag,Values).

hdrug_menu_flag(Flag) :-
    hdrug_menu_flag(Flag,[]).

hdrug_menu_flag(Flag,Values0) :-
	flag(Flag,Value),
	(   member(Value,Values0)
	->  Values0=Values
	;   Values=[Value|Values0]
	),
	tcl("set flag(~w) ~w",[Flag,Value]),
	menu_flag('.menu.hdrug.m',Flag,Values).

menu_flag(Menu,Flag,Values) :-
	tcl('~w add cascade -label "~w" -menu ~w.~w',
	    [Menu,Flag,Menu,Flag]),
	tcl('menu ~w.~w',[Menu,Flag],SubMenu),
	menu_flag_values(Values,Flag,SubMenu).

menu_flag_values([],Flag,Menu) :-
	tcl('~w add command -label Help -command "prolog hdrug_gui:tk_help_flag(~w)"',
	    [Menu,Flag]).
menu_flag_values([V0|V],Flag,Menu) :-
	tcl('~w add radiobutton -label ~w -value ~w -variable flag(~w) \
	      -command "prolog flag(~w,_,~w) "',
	      [Menu,V0,V0,Flag,Flag,V0]),
	menu_flag_values(V,Flag,Menu).

tk_init :-
	tk_new([top_level_events]).

:- public tk_help_flag/1.
tk_help_flag(Flag) :-
    (	user:help_flag(Flag,Text)
    ->  tcl("tk_dialog .dialog {Help on ~w} {~s} {} 0 OK",[Flag,Text])
    ;	tcl("tk_dialog .dialog {Help} {No help available for ~w} {} 0 OK",
	    [Flag])
    ).
%% used by hdrug.tcl

:- public send_all/2.
send_all(Term,TclVar):-
    atom_term(Term,f(Var,Call)),
    send_all(Var,Call,TclVar).

%% used by hdrug.tcl
send_all(Var,Call,TclVar):-
    findall(Var,hook(user:Call),Vars0),
    wrap_write(Vars0,Vars),
    tcl_eval([set,TclVar,br(Vars)]).

wrap_write([],[]).
wrap_write([H|T0],[write(H)|T]) :-
    wrap_write(T0,T).

start_tkconsol :-
    flag(tkconsol,Val),
    start_tkconsol(Val).

start_tkconsol(off).
start_tkconsol(undefined).
start_tkconsol(on) :-
    tcl('frame .cons
         text .cons.console -width 80 -height 10 -yscrollcommand ".cons.sb set"
         bind Text <Control-u> {
             tkTextSetCursor %W {insert linestart}
         if {[%W compare insert == {insert lineend}]} {
             %W delete insert
         } else {
             %W delete insert {insert lineend}
         }
	 }
         scrollbar .cons.sb -command ".cons.console yview"
         pack .cons.sb -side left -fill both
         pack .cons.console -fill both -expand 1  -side left
         pack .cons -fill both  -after .t
',[]),
    flag(tcl_interp,Interp),
    (	on_exception(Error,tk_terminal(Interp,".cons.console",InS,OutS,ErrS),
		     (	 print_message(error,Error),
			 fail
		     )
		    )
    ->	prolog_flag(user_input, InOld, InS),
	prolog_flag(user_output, OutOld, OutS),
	prolog_flag(user_error, ErrOld, ErrS),
	set_flag(tkconsol_streams,old(InOld,OutOld,ErrOld)),
	set_input(InS),
	set_output(OutS),
	format(user_error,"TkConsol started (experimental)~n",[]),
	tcl('.menu.file.m add separator'),
	tcl('.menu.file.m add command -label "Enlarge Console (1 line)" \
	   -command { .cons.console configure -height \
             [expr [lindex [.cons.console configure -height] 4]+1]
	   }',[]),
	tcl('.menu.file.m add command -label "Enlarge Console (5 lines)" \
	   -command { .cons.console configure -height \
             [expr [lindex [.cons.console configure -height] 4]+5]
	   }',[]),
	tcl('.menu.file.m add command -label "Shrink Console (1 line)" \
	   -command { .cons.console configure -height \
             [expr [lindex [.cons.console configure -height] 4]-1]
	   }',[]),
	tcl('.menu.file.m add command -label "Shrink Console (5 lines)" \
	   -command { .cons.console configure -height \
             [expr [lindex [.cons.console configure -height] 4]-5]
	   }',[])
        
    ;   tcl('destroy .cons'),
	format(user_error,"ERROR: Cannot start tkconsol~n",[])
    ).

tkconsol_larger :-
    tkconsol_larger(1).

tkconsol_larger(I) :-
    integer(I),
    tcl('.cons.console configure -height \
             [expr [lindex [.cons.console configure -height] 4]+~w]',[I]).

tkconsol_smaller :-
    tkconsol_smaller(1).

tkconsol_smaller(I) :-
    integer(I),
    tcl('.cons.console configure -height \
             [expr [lindex [.cons.console configure -height] 4]-~w]',[I]).

reset_tkconsol :-
    flag(tkconsol_streams,Old,undefined),
    reset_tkconsol(Old).

reset_tkconsol(undefined).
reset_tkconsol(old(In,Out,Err)):-
    prolog_flag(user_input,_,In),
    prolog_flag(user_output,_,Out),
    prolog_flag(user_error,_,Err),
    set_input(In),
    set_output(Out).


%% TODO: do something special for those numbers which not currently have
%% an obj button, e.g., because > 39.
notify_active_obj(No) :-
    if_gui(notify_active_obj0(No)).

notify_active_obj0(No) :-
    flag(active_obj,No0,No),
    (	No0 == undefined
    ->  true
    ;   tcl('catch {.bb.o.obj~w configure -background [option get .bb.o.obj~w background Background]}',[No0,No0])
    ),
    tcl('if {![winfo exists .bb.o.obj~w]} { hdrug_object ~w } ',[No,No]),
    tcl('catch {.bb.o.obj~w configure -background [option get .bb.o.obj~w highlightBackground HighlightBackground]}',[No,No]).

%tk_help_message(Class,Key) :-
%    gen_sym(W,'.dt'),
%    user:help_info(Class,Key,Usage,Expl),
%    tcl("display_text ~w {Help on ~w ~w} {~s} {~s}",
%	[W,Class,Key,Expl,Usage]).

:- public spy_atom/1, nospy_atom/1.
spy_atom(Atom) :-
    atom_term(Atom,Term),
    user:spy(Term).

nospy_atom(Atom) :-
    atom_term(Atom,Term),
    user:nospy(Term).

tcl_interp(Val):-
    flag(tcl_interp,Val),
    (	Val==undefined
    ->  raise_exception(hdrug_error('This command assumes the graphical user interface is running~n',[]))
    ;   true
    ).


atmost_prefix(Max,Values0,Values) :-
    (	Max =:= 0
    ->	Values0=Values
    ;   length(Values0,Len),
	(   Len > Max
	->  length(Values,Max),
	    lists:append(Values,_,Values0)
	;   Values0=Values
	)
    ).

findall_atmost(Max,V,Goal,Vs) :-
    statistics(walltime,[Ref,_]),
    bb_put(sol:Ref,1),
    call_cleanup(findall_atmost_(Ref,V,Goal,Vs,Max),
		 bb_delete(sol:Ref,_)).

:- dynamic sol/2.
findall_atmost_(Ref,V,Goal,Vs,Number):-
    call(Goal),
    assertz(sol(Ref,V)),
    more_or_stop(Ref,Vs,Number),
    !.
findall_atmost_(Ref,_,_,Vs,_):-  %% never got to the maximum
    findall(V,retract(sol(Ref,V)),Vs).
    
more_or_stop(Ref,Vs,Number):-
    bb_get(sol:Ref,N),
    (	N < Number
    ->	N1 is N+1,
	bb_put(sol:Ref,N1),
	fail
    ;   findall(V,retract(sol(Ref,V)),Vs)
    ).

%display_sentence_key(Sent) :-
%    user:a_sentence(Key,Sent),
%    format(user_error,"key: ~w~n",[Key]).

gui_user_confirmation(Msg,Args) :-
    charsio:format_to_chars(Msg,Args,Chars),
    tcl("tk_dialog .dialog {Confirmation} {~s} {} 0 NO YES",[Chars],Return),
    Return='1'.

