.TH hdrug 1
.SH NAME
.B hdrug

.SH SYNOPSIS
.B hdrug <options>

.SH DESCRIPTION

Hdrug is a graphical user environment for natural language processing
in Prolog. The environment facilitates the development of logic
grammars and associated interpreters such as parsers and
generators. The system offers a graphical user interface (based on
Tcl/Tk), a command interpreter, various visualization tools, and
various useful libraries (such as a library to compile feature
equations into Prolog terms). 

Hdrug is implemented in SICStus Prolog.

Further documentation is available as
http://www.let.rug.nl/~vannoord/Hdrug/Manual/

.SH OPTIONS

.TP 8
.B -notk
don't start graphical user interface. This option is equivalent to the 
option tcltk=off.

.TP 8
.B -tk     
starts the graphical user interface (default). This option is
equivalent to the option tcltk=on.

.TP 8
.B Att=Val
sets the global variable Att to value Val. Val is parsed as a Prolog
term. Refer to the section on global variables below.

.TP 8
.B -flag Att Val
sets the global variable Att to value Val. Val is parsed as a Prolog
atom. Refer to the section on global variables below.

.TP 8
.B -pflag Att Val
sets the global variable Att to value Val. Val is parsed as a Prolog
atom. 

.TP 8
.B -cmd Cmd
executes the Prolog goal Cmd. Cmd is parsed as a Prolog term. It
should end with a dot. Example: hdrug -notk -cmd "(lising,halt)."

.TP 8
.B -l File
loads File, using use_module.

.TP 8
.B
-parser Parser on/off
Declares Parser to be a parser, and sets its status to on or off.
Refer to the section on parsers and generators below.

.TP 8
.B 
-generator Generator on/off
Declares Parser to be a parser, and sets its status to on or off.
Refer to the section on parsers and generators below.

.TP 8
.B
-help
Shows options in summary, and quits.

.TP 8
.B
-dir Dir
adds directory Dir to Prolog's library directories list.

Apart from these options, Hdrug applications can extend the list of
possible startup options by adding definitions to the multifile
predicate option/3. Short usage information for such options can be
defined with further definitions for the multifile predicate
usage_option/1.  

.SH Global Variables

.TP 8
.B parser <Atom>
Determines which parser is currently the parser to use for
parse commands.

.TP 8
.B top_features: <Atom>
Determines top category used by the parsers. If the value is `vp',
then the predicate user:top(vp,Cat) determines the top category (start
symbol).

.TP 8
.B useful_try_check: [on,off]
determines during parser comparison whether a sentence
should be parsed with a parser which timed-out on a shorter
sentence.
Default:off

.TP 8
.B parser(Parser): [on,off]
determines whether Parser is currently active or not, i.e. whether it
will take part in parser comparison runs or not. 

.TP 8
.B generator(Generator): [on,off]
similar to  parser(Parser), but for generators.

.TP 8
.B generator: <Atom>
Current generator. 

.TP 8
.B object_exists_check: [on,off]
Determines whether the system should check if an object that
results from a parse does already exist.
Default:on

.TP 8
.B application_name: <Atom>
Used on the Tcl/Tk side. Determines which applicaiton default
file is loaded, and the title of the widget.

.TP 8
.B tcltk: [on,off]
Determines whether the X environment starts.
Default: on

.TP 8
.B tconsol: [on,off]
Determines whether or not the tkconsol feature (cf. library(tkconsol)
should be used. Default: off.

.TP 8
.B debug: [0,1,2]
Determines the number and detail of continuation messages. 
Default: 0 (minimum)

.TP 8
.B object_saving: [on,off,semi]
If on, then objects are always saved. If off, then objects are never
saved. If semi, then all existing objects are removed if a new parse
is attempted. 

.TP 8
.B width <Integer>
Used to indicate the width of the tty, for tty-based output of trees.

.TP 8
.B current_no <Integer>
Used internally to count the current number of object.

.TP 8
.B found_solutions <Integer>
Used internally to count the number of solutions found sofar.

.TP 8
.B demo [on,off]
If on, some more information after succesful parse is written to tty.

.TP 8
.B max_objects <Integer>
Maximal number of objects for a single parse. Parser will terminate if 
more parse results could be found.

.TP 8
.B print_table_total [on,off]
determines whether during a parse comparison totals should be
displayed after each sentence.

.TP 8
.B blt_graph_lines [on,off]
should we connect dots in a blt_graph widget?
Default: off.

.TP 8
.B node_skip <Integer>
used for LaTeX output of trees (pstree).  Determines vertical distance
of nodes in centimeters. 

.TP 8
.B start_results_within_bound
Determines for the results_within_bound command (st_tty and st_blt) at
which number of millisecond reporting should start. Default: 100.

.TP 8
.B end_results_within_bound
Determines for the results_within_bound command (st_tty and st_blt) at
which number of millisecond reporting should end. Default: 5000.

.TP 8
.B incr_results_within_bound
Determines for the results_within_bound command (st_tty and st_blt) with
which number of millisecond increment percentages should be shown. 
Default: 100.

.SH COPYRIGHT
Copyright \(co 1997 by Gertjan van Noord, RUG.

This program is distributed
under Gnu General Public License (cf. the file COPYING in
distribution).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

.SH COPYRIGHT
Copyright \(co 1992-1998 by Gertjan van Noord, RUG.
This program is distributed
under Gnu General Public License (cf. the file COPYING in
distribution).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; version 2 of the License.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


.SH AUTHOR
Gertjan van Noord, vannoord@let.rug.nl

